/*
 * code file for wmbatppc
 * by Lou
 */


#include "wmbatppc.h"

#include "wmgeneral.h"

#include "wmbatppc-master.xpm"

pmud_info sys_pmu ;

char wmbatppc_mask_bits[64*64];
int wmbatppc_mask_width = 64;
int wmbatppc_mask_height = 64;

int update_rate=100000;

char *ProgName;

time_t curtime;
time_t prevtime;

int mode = 0 ; // default: no info
int screen = 0;  // default: Quality screen is displayed

void showPlug (void){
	copyXPMArea(158,36,12,16,2,2) ;
}

void hidePlug (void){
	copyXPMArea(158,53,12,16,2,2) ;
}

void showCharg (void){
	copyXPMArea(134,36,12,16,50,2) ;
}

void hideCharg (void){
	copyXPMArea(134,53,12,16,50,2) ;
}

void drawBatteryState(int bat_side, int bat_state){
	int x, y, dest_x ;

	if (bat_state == NO_BATT){
		x = 108 + 27 * bat_side ;
		y = 85 ;
	} else {
		x = 146 + 27 * bat_state ;
		y = 6 + 15 * bat_side ;
	}

	dest_x = 6 + 27 * bat_side ;

	copyXPMArea (x,y,25,13,dest_x,47) ; 
}

void drawJukeBox (int state){
	copyXPMArea (173,36+32*state,57,31,4,13) ;
}


int open_pmud_socket() {

     	int ret;

	     struct sockaddr_in sin ;
	     sin.sin_family = PF_INET ;
	     sin.sin_port = htons (PMUD_PORT) ;
	     sin.sin_addr.s_addr = htonl (INADDR_LOOPBACK) ;

	  if ((ret = socket (PF_INET, SOCK_STREAM, 0)) >= 0) {
	       if (connect (ret, (struct sockaddr *) &sin, sizeof (sin)) >= 0) {
		    return ret;
	       } else {
		    close (ret);
	       }
	  }
     	return -1;
}


  /*

# On the lombard, the line we get from pmud is
#   S {flags charge maxcharge ibat vbat} {(same for R batt)}
#
# Flags is a set of 3 bits, in binary, which we number left-to-right
# from 0 to 2.  (Flag 0 is 1 in the byte we get from the PMU.)
# Flag 0 indicates that the AC is connected.
# Flag 1 indicates that the battery is being charged.
# Flag 2 indicates that the battery is present.
#
# The vbat field gives the battery voltage in millivolts maybe.
# The ibat field gives the battery current, positive when charging.
  
*/

enum { KL_IBOOK, KL_PISMO, KL_UNKNOWN };

static int keylargo_identify ()
{
     FILE *fd;
     char buf[25];
     if (!(fd = fopen ("/proc/device-tree/pci/mac-io/media-bay/compatible", "ro")))
	  return KL_IBOOK;
     /* no media-bay.  definately an iBook...
      * or a Pismo running a kernel without OF-devtree support...
      */
     fscanf (fd, "%s", buf);
     fclose (fd);
     if (!strcmp ("keylargo-media-bay", buf))	// only the pismo should have one
	  return KL_PISMO;	// has one, its a pismo
     else
	  return KL_UNKNOWN;
     /* has a media bay, IDs as a Keylargo, but not a
      * "keylargo-media-bay.... what the...
      */
}



void read_g3_pmu ()
{	/* smart battery systems are the same */

     unsigned char par;
     int current, charge, maxcharge, ac_hold = -1;
     char *b;
     char c[150];
     static int sock = -1;

     
     if (sock < 0) {
	  if ((sock = open_pmud_socket ()) < 0) {
	       sys_pmu.time_left = -2;
	       return;
	  } else {
	       read (sock, c, 150);
	  }
     }
#ifdef PMUD_REFRESH
     read (sock, c, 150);
     shutdown (sock, SHUT_RDWR);
     close (sock);
     sock = -1;
#else
     write (sock, "\n", 1);
     read (sock, c, 150);
#endif
     b = strtok (c, PMUD_INPUT_DELIM);
     if (b[0] != 'S')
	  return;
     sys_pmu.batt_state[LEFT_BATTERY] = NO_BATT ;
     sys_pmu.batt_state[RIGHT_BATTERY] = NO_BATT ;

     sys_pmu.time_left = 0;
     for (par = 0; par < 2; par++) {
	  ac_hold = atoi (strtok (0, PMUD_INPUT_DELIM));
	  if (ac_hold % 10) {
	       sys_pmu.b[par].available = IS_BATT ;
	       sys_pmu.b[par].charging = (((ac_hold / 10) % 10) == 1);
	       charge = atoi (strtok (0, PMUD_INPUT_DELIM));
	       maxcharge = atoi (strtok (0, PMUD_INPUT_DELIM)) - 2 ;
	       sys_pmu.b[par].current = atoi (strtok (0, PMUD_INPUT_DELIM));
	       sys_pmu.b[par].voltage = atoi (strtok (0, PMUD_INPUT_DELIM));
	       sys_pmu.b[par].percentage = charge * 100 / maxcharge ;
	       sys_pmu.time_left += ((ac_hold / 100) ? maxcharge - charge : charge);
	       sys_pmu.batt_state[par] = IS_BATT ;
	  } else {
	       sys_pmu.b[par].available = NO_BATT ;
	       sys_pmu.b[par].percentage = 0;
	       sys_pmu.b[par].current = 0;
	       sys_pmu.b[par].voltage = 0.0;
	       sys_pmu.b[par].charging = 0;
	  }
     }
     sys_pmu.ac_connected = ac_hold / 100;
     current = sys_pmu.b[LEFT_BATTERY].current + sys_pmu.b[RIGHT_BATTERY].current;
     if (current < 0)
	  sys_pmu.time_left = sys_pmu.time_left * 3552 / (current * -60);
     else if (sys_pmu.show_charge_time && (current > 0))
	  sys_pmu.time_left = sys_pmu.time_left * 3552 / (current * 60);
     else
	  sys_pmu.time_left = -1;
}



void DisplayBat(void) {
	
	int keylargo ;
	unsigned int par = 0 ;
	int hour = 0 ;
	int min = 0 ;
	int volts = 0 ;
	int millivolts = 0 ;
	int i;

	sys_pmu.show_charge_time = 1 ;

	keylargo = keylargo_identify() ; // TODO : use this detection

	read_g3_pmu () ;
	
	if (sys_pmu.time_left == -1)
		drawJukeBox(INFINITE) ;
	else
	{	drawJukeBox(LIMITED) ;
		hour =	sys_pmu.time_left / 60 ;
		min = sys_pmu.time_left % 60 ;
		// printf("%i,%d,%d\n",sys_pmu.time_left,hour,min);
		BlitNum(hour,15,31,BIG,1);  
		BlitNum(min,33,31,BIG,1);  
	}

	if (sys_pmu.ac_connected)
		showPlug() ;
	else 
		hidePlug() ;

	if (sys_pmu.b[LEFT_BATTERY].charging || sys_pmu.b[RIGHT_BATTERY].charging)
		showCharg() ;
	else
		hideCharg() ;

     for (par=0; par<2; par++){
	if (sys_pmu.b[par].available == NO_BATT){
		drawBatteryState(par,NO_BATT) ;
	}else {
		if (sys_pmu.b[par].percentage > 70)
			drawBatteryState(par,HIGH_BATT) ;
		else if (sys_pmu.b[par].percentage > 25)
			drawBatteryState(par,MED_BATT) ;
		else if ((sys_pmu.b[par].percentage > 5) || (sys_pmu.b[par].charging) || (sys_pmu.b[1-par].percentage > 5))
			drawBatteryState(par,LOW_BATT) ;
		else 
			{  
			for (i=0; i<update_rate/50000; i++){
				drawBatteryState(par,NO_BATT) ;
				RedrawWindow() ;
				usleep(200000) ;
				drawBatteryState(par,LOW_BATT) ;
				RedrawWindow() ;
			}
		}

		if (sys_pmu.b[par].percentage == 100){
			BlitNum(sys_pmu.b[par].percentage,9+par*25,49,SMALL,0) ;
		} else {
			BlitNum(sys_pmu.b[par].percentage,11+par*25,49,SMALL,1) ;
		}
	
		volts = sys_pmu.b[par].voltage / 1000 ;
		millivolts =  sys_pmu.b[par].voltage % 1000 ;
		BlitNum(volts,15,2,SMALL,1) ;
		BlitNum(millivolts,29,2,SMALL,0) ;
		copyXPMArea(74,69,2,3,27,7);
	}
     }

	if (sys_pmu.b[LEFT_BATTERY].available == NO_BATT 
			&& sys_pmu.b[RIGHT_BATTERY].available == NO_BATT)
		copyXPMArea(1,84,34,10,15,2) ;
}

/* SIGCHLD handler */
void sig_chld(int signo)
{
  waitpid((pid_t) -1, NULL, WNOHANG);
  signal(SIGCHLD, sig_chld);
}

int main(int argc, char *argv[]) {
  int i;
  
  /* Parse Command Line */
  
  signal(SIGCHLD, sig_chld);
  ProgName = argv[0];
  if (strlen(ProgName) >= 5)
    ProgName += (strlen(ProgName) - 5);
  
  for (i=1; i<argc; i++) {
    char *arg = argv[i];
    
    if (*arg=='-') {
      switch (arg[1]) {
      case 'd' :
	if (strcmp(arg+1, "display")) {
	  usage();
	  exit(1);
	}
	break;
      case 'g' :
	if (strcmp(arg+1, "geometry")) {
	  usage();
	  exit(1);
	}
	break;
      case 'v' :
	printversion();
	exit(0);
	break;
      case 'r':
	if (argc > (i+1)) {
	  update_rate = (atoi(argv[i+1]) * 1000);
	  i++;
	}
	break;
      default:
	usage();
	exit(0);
	break;
      }
    }
  }
  
  wmbatppc_routine(argc, argv);
  
  return 0;
}

/*
 * Main loop
 */
void wmbatppc_routine(int argc, char **argv) {
  XEvent Event;
  struct timeval tv={0,0};
  struct timeval last={0,0};
  
  createXBMfromXPM(wmbatppc_mask_bits, wmbatppc_master_xpm, wmbatppc_mask_width, wmbatppc_mask_height);
  
  openXwindow(argc, argv, wmbatppc_master_xpm, wmbatppc_mask_bits, wmbatppc_mask_width, wmbatppc_mask_height);
  
  RedrawWindow();
  
  
  while (1) {
    
    curtime = time(0);
    
    if (1) {
      memcpy(&last, &tv, sizeof(tv));
      
      /*
       * Update display
       */

      DisplayBat();
      
      RedrawWindow();
    }
    
    /*
     * X Events
     */
    while (XPending(display)) {
      XNextEvent(display, &Event);
      switch (Event.type) {
      case Expose:
	RedrawWindow();
	break;
      case DestroyNotify:
	XCloseDisplay(display);
	exit(0);
      case ButtonPress:
	switch (screen) {
	case 0: screen=1; break;
	case 1: screen=0; break;
	};
	break;
      }
    }
    
    usleep(update_rate);
  }
}

/*
 * Blits a string at given co-ordinates
 */
void BlitString(char *name, int x, int y, int is_big) {
  int	i;
  int	c;
  int	k;

  k = x;

  for (i=0; name[i]; i++){
	c = toupper(name[i]); 
	if (c >= 'A' && c <= 'Z'){   // its a letter
		c -= 'A';
		copyXPMArea(c * 6, 74, 6, 8, k, y);
		k += 6;
	} else {
	  	if (c>='0' && c<='9'){   // its a number or symbol
	 		c -= '0';
	  		if (is_big == BIG){
	 			copyXPMArea(68 + c * 7 , 22, 7, 10, k, y);
				k += 7 ;
			} else {
	 			copyXPMArea(c * 6, 64, 6, 8, k, y);
	  			k += 6;
			}
	  	} else {
	    		copyXPMArea(1, 84, 6, 8, k, y);
	    		k += 6;
	  	}
	}
   }
}

void BlitNum(int num, int x, int y, int is_big, int two_digits) {
  char buf[1024];
  int newx=x;
 
  if (two_digits) 
  	sprintf(buf, "%02i", num);
  else 
  	sprintf(buf, "%03i", num);
  
  BlitString(buf, newx, y, is_big);
}

/*
 * Usage
 */
void usage(void) {
  fprintf(stderr, "\nwmbatppc - Lou <titelou@free.fr>  http://titelou.free.fr\n\n");
  fprintf(stderr, "usage:\n");
  fprintf(stderr, "    -display <display name>\n");
  fprintf(stderr, "    -r	update rate in milliseconds (default:100)\n");
  fprintf(stderr, "\n");
}

/*
 * printversion
 */
void printversion(void) {
  fprintf(stderr, "wmbatppc v%s\n", WMBATPPC_VERSION);
}
